from typing import TypedDict, Optional


class Notification(TypedDict):
    """
    Notification
    https://firebase.google.com/docs/reference/fcm/rest/v1/projects.messages#Notification
    """
    title: str
    body: str
    image: str


class LightSettings(TypedDict):
    """
    Light Settings
    https://firebase.google.com/docs/reference/fcm/rest/v1/projects.messages#LightSettings
    """
    color: dict
    light_on_duration: str
    light_off_duration: str


class AndroidNotification(TypedDict):
    """
    Android Notification
    https://firebase.google.com/docs/reference/fcm/rest/v1/projects.messages#AndroidNotification
    """
    title: str
    body: str
    icon: str
    color: str
    sound: str
    tag: str
    click_action: str
    body_loc_key: str
    body_loc_args: list[str]
    title_loc_key: str
    title_loc_args: list[str]
    channel_id: str
    ticker: str
    sticky: bool
    event_time: str
    local_only: bool
    notification_priority: str
    default_sound: bool
    default_vibrate_timings: bool
    default_light_settings: bool
    vibrate_timings: list[str]
    visibility: str
    notification_count: int
    light_settings: LightSettings
    image: str


class AndroidFcmOptions(TypedDict):
    """
    Android FCM Options
    https://firebase.google.com/docs/reference/fcm/rest/v1/projects.messages#AndroidFcmOptions
    """
    analytics_label: str


class AndroidConfig(TypedDict):
    """
    Android Config
    https://firebase.google.com/docs/reference/fcm/rest/v1/projects.messages#AndroidConfig
    """
    collapse_key: str
    priority: str
    ttl: str
    restricted_package_name: str
    data: dict[str, str]
    notification: AndroidNotification
    fcm_options: AndroidFcmOptions
    direct_boot_ok: bool


class WebpushFcmOptions(TypedDict):
    """
    Webpush FCM Options
    https://firebase.google.com/docs/reference/fcm/rest/v1/projects.messages#WebpushFcmOptions
    """
    link: str
    analytics_label: str


class WebpushConfig(TypedDict):
    """
    Webpush Config
    https://firebase.google.com/docs/reference/fcm/rest/v1/projects.messages#WebpushConfig
    """
    headers: dict[str, str]
    data: dict[str, str]
    notification: dict
    fcm_options: WebpushFcmOptions


class ApnsFcmOptions(TypedDict):
    """
    APNS FCM Options
    https://firebase.google.com/docs/reference/fcm/rest/v1/projects.messages#ApnsFcmOptions
    """
    analytics_label: str
    image: str


class ApnsConfig(TypedDict):
    """
    APNS Config
    https://firebase.google.com/docs/reference/fcm/rest/v1/projects.messages#ApnsConfig
    """
    headers: dict[str, str]
    payload: dict
    fcm_options: ApnsFcmOptions


class FcmOptions(TypedDict):
    """
    FCM Options
    https://firebase.google.com/docs/reference/fcm/rest/v1/projects.messages#FcmOptions
    """
    analytics_label: str


class Message(TypedDict):
    """
    FCM Message Resource
    https://firebase.google.com/docs/reference/fcm/rest/v1/projects.messages
    """
    name: Optional[str]
    data: Optional[dict[str, str]]
    notification: Optional[Notification]
    android: Optional[AndroidConfig]
    webpush: Optional[WebpushConfig]
    apns: Optional[ApnsConfig]
    fcm_options: Optional[FcmOptions]
    token: str
    topic: Optional[str]
    condition: Optional[str]



import time
import json
import google.auth.transport.requests
from google.oauth2 import service_account
from typing import Union, Iterable
from .errors import InvalidCredentialsError


class PyFCMAuthenticator:
    def __init__(self):
        """PyFCMAuth Initialization"""
        self.scopes: Iterable[str] = ('https://www.googleapis.com/auth/firebase.messaging',)
        self.latest_refresh_time = 0
        self.auto_refresh = None
        self._credentials = None

    @property
    def project_id(self):
        """Get project id from credentials"""
        if not self._credentials:
            raise InvalidCredentialsError("credentials_initialization_required")
        return self._credentials.project_id

    @property
    def access_token(self):
        """Get access token from credentials"""
        if not self._credentials:
            raise InvalidCredentialsError("credentials_initialization_required")
        if self.auto_refresh:
            self.refresh_credentials()
        return self._credentials.token

    def refresh_credentials(self):
        """Refresh credentials"""
        now_time = time.time()
        if now_time - self.latest_refresh_time > 1800:
            request = google.auth.transport.requests.Request()
            self._credentials.refresh(request)
            self.latest_refresh_time = now_time

    def init_credentials_from_file(
        self,
        credentials_file_path: str,
        auto_refresh: bool = True
    ):
        """
        Initialize credentials from file

        :param credentials_file_path: Path to the json file containing the credentials.
        :param auto_refresh: Decide whether to automatically refresh the Google Access Token.
          - True (Default): The Access Token is checked immediately before sending the message,
                            and is automatically renewed 30 minutes before expiration.
          - False: Access Token is not refreshed automatically. (Suitable for short-term use.)
        """
        self._credentials = service_account.Credentials.from_service_account_file(
            filename=credentials_file_path,
            scopes=self.scopes
        )
        self.auto_refresh = auto_refresh
        self.refresh_credentials()

    def init_credentials(
        self,
        credentials: Union[str, dict],
        auto_refresh: bool = True
    ):
        """
        Initialize credentials from json string or dict

        :param credentials: Json string or dictionary containing the credentials.
        :param auto_refresh: Decide whether to automatically refresh the Google Access Token.
          - True (Default): The Access Token is checked immediately before sending the message,
                            and is automatically renewed 30 minutes before expiration.
          - False: Access Token is not refreshed automatically. (Suitable for short-term use.)
        """
        self._credentials = service_account.Credentials.from_service_account_info(
            info=(
                credentials
                if isinstance(credentials, dict)
                else json.loads(credentials)
            ),
            scopes=self.scopes
        )
        self.auto_refresh = auto_refresh
        self.refresh_credentials()



import aiohttp
from typing import Union
from .pyfcm_auth import PyFCMAuthenticator
from .message import Message
from .errors import (
    UnspecifiedError, InvalidArgumentError, UnregisteredError,
    SenderIdMismatchError, QuotaExceededError, UnavailableError,
    InternalServerError, ThirdPartyAuthError
)


class AioPyFCM(PyFCMAuthenticator):

    async def __aenter__(self):
        self.session = aiohttp.ClientSession()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.session.close()

    def __init__(self):
        """AioPyFCM Initialization"""
        super().__init__()
        self.session = None
        self.endpoint = "https://fcm.googleapis.com"

    async def send(self, message: Message):
        """
        Send message to FCM server

        :param message: The message to send.
        :return: JSON response from the FCM Server.
        """
        if not self.session:
            async with aiohttp.ClientSession() as session:
                self.session = session
                return await self._post_message(message)
        else:
            return await self._post_message(message)

    async def _post_message(self, message: Message):
        """
        Internal method to post a message.
        """
        async with self.session.post(
            url=f"{self.endpoint}/v1/projects/{self.project_id}/messages:send",
            headers={
                "Authorization": f"Bearer {self.access_token}",
                "Content-Type": "application/json"
            },
            json={"message": message}
        ) as response:
            result = await response.json()
            if response.status == 200:
                return result
            elif response.status == 400:
                raise InvalidArgumentError(result)
            elif response.status == 401:
                raise ThirdPartyAuthError(result)
            elif response.status == 403:
                raise SenderIdMismatchError(result)
            elif response.status == 404:
                raise UnregisteredError(result)
            elif response.status == 429:
                raise QuotaExceededError(result)
            elif response.status == 500:
                raise InternalServerError(result)
            elif response.status == 503:
                raise UnavailableError(result)
            else:
                raise UnspecifiedError(result)
